from .prior import MeansGaussianPrior, MeansUniformPrior, CovarsStaticPrior, \
                   DiagCovarsUniformPrior, WeightsUniformPrior,\
                   WeightsDirichletPrior, WeightsStaticPrior, GMMPrior
